let split_int =
	let rec aux = fun
		| [] 0 -> [0]
		| l  0 -> l
		| l  n -> aux ((n mod 10) :: l) (n / 10)
	in
	aux [];;

(* ======================== PARTIE 1 ======================== *)

type ('a, 'b) automaton =
{
	starts : 'a list;
	accepts : 'a list;
	trans : 'a -> 'b -> 'a list;
};;

let gen_aut_length_div n =
	{ starts = [0]; accepts = [0]; trans = fun s _ -> [(s+1) mod n] };;

let gen_aut_div n =
	{ starts = [0]; accepts = [0]; trans = fun s i -> [(s * 10 + i) mod n] };;

let run aut input =
	let rec aux s = function
		| [] -> (intersect aut.accepts s) <> []
		| h :: t -> aux (flat_map (fun x -> aut.trans x h) s) t
	in
	aux aut.starts input;;

let aut = gen_aut_div 7 in
let test n = printf__printf "%d - %b\n" n (run aut (split_int n)) in
	test 0;
	test 1;
	test 6;
	test 7;
	test 14;
	test 15;
	test 5040;;

let rec make_unique l =
	let rec remove x = function
		| [] -> []
		| h :: t when h = x -> remove x t
		| h :: t -> h :: remove x t
	in
	match l with
		| [] -> []
		| h :: t -> h :: make_unique(remove h t);;

let run aut input =
	let rec aux s = function
		| [] -> (intersect aut.accepts s) <> []
		| h :: t -> aux (make_unique(flat_map (fun x -> aut.trans x h) s)) t
	in
	aux aut.starts input;;

let aut = gen_aut_div 7 in
let test n = printf__printf "%d - %b\n" n (run aut (split_int n)) in
	test 0;
	test 1;
	test 6;
	test 7;
	test 14;
	test 15;
	test 5040;;
    
(* ======================== PARTIE 2 ======================== *)

type ('a, 'b) automaton =
{
	starts : 'a list;
	accepts : 'a list;
	trans : (('a * 'b) * 'a) list;
};;

let gen_aut_div n =
	let trans = ref [] in
	for state = 0 to n-1 do
		for digit = 0 to 9 do
			trans := ((state ,digit),(state * 10 + digit) mod n) :: !trans;
		done;
	done;
	{ starts = [0]; accepts = [0]; trans = !trans; };;
	
let gen_aut_length_div n =
	let trans = ref [] in
	for state = 0 to n-1 do
		for digit = 0 to 9 do
			trans := ((state ,digit),(state + 1) mod n) :: !trans;
		done;
	done;
	{ starts = [0]; accepts = [0]; trans = !trans; };;


let run aut w =
	(* tat courant et mot restant *)
	let cur = ref (hd aut.starts)
	and rem = ref w in
	(* tant qu'il reste des choses  lire *)
	try
		while !rem <> [] do
		(* l'exception est rattrape en dehors de la boucle si la transition n'existe pas *)
			cur := assoc (!cur, hd !rem) aut.trans;
			rem := tl !rem;
		done;
		mem !cur aut.accepts
	with
		| Not_found -> false;;

let aut = gen_aut_div 7 in
let test n = printf__printf "%d - %b\n" n (run aut (split_int n)) in
	test 0;
	test 1;
	test 6;
	test 7;
	test 14;
	test 15;
	test 5040;;
    
let run aut word =
	(* on gnralise  l'acceptation d'un mot  partir d'un tat *)
	let rec aux state = function
		| [] -> mem state aut.accepts (* vrifie qu'on est dans un tat final *)
		| c::w -> aux (assoc (state, c) aut.trans) w
	in
	try
		(* on part de l'tat initial *)
		aux (hd aut.starts) word
	with
		| Not_found -> false;;

let aut = gen_aut_div 7 in
let test n = printf__printf "%d - %b\n" n (run aut (split_int n)) in
	test 0;
	test 1;
	test 6;
	test 7;
	test 14;
	test 15;
	test 5040;;

(*
 * Pour chaque transition ((x, c'), y) si c=c' et x est dans set, on
 * renvoie [y], sinon [].
 * On fait ensuite la concatnation du tout et on unifie
 *)
let rec multi_trans aut c set =
	make_unique (flat_map (function ((x, c'), y) -> if c = c' && mem x set then [y] else []) aut.trans);;

(*
 * fonctions rcursive sur la liste des transitions
 *)
let auto_alphabet aut =
	let rec alphabet = function
		| [] -> []
		| ((_, c), _) :: q -> let alpha = alphabet q in
				if mem c alpha then alpha else c :: alpha
	in
	alphabet aut.trans;;



    

